#include <gtc/matrix_transform.hpp>
#define _USE_MATH_DEFINES
#include <math.h>
#include "Camera.h"

Camera::Camera(glm::vec3 position) : position(position) {
	up = glm::vec3(0.0f, 1.0f, 0.0f);
	center = glm::vec3(position.x + 10.0f, position.y, position.z);
	velocity = 0.5f;
    frustumPlanes.reserve(6);
    
}

Camera::~Camera() {

}

void Camera::move(const Uint8* currentKeyStates) {
	glm::vec3 direction = glm::vec3(center.x, 0.0f, center.z) - glm::vec3(position.x, 0.0f, position.z);
	direction = glm::normalize(direction);
	glm::vec3 directionNormal = glm::vec3(direction.z, 0.0f, -direction.x);
	glm::vec3 frontBackStep = direction * velocity;
	glm::vec3 leftRightStep = directionNormal * velocity;
	if (currentKeyStates[SDL_SCANCODE_W]) {
		position += frontBackStep;
		center += frontBackStep;
	}
	if (currentKeyStates[SDL_SCANCODE_S]){
		position -= frontBackStep;
		center -= frontBackStep;
	}
	if (currentKeyStates[SDL_SCANCODE_A]) {
		position += leftRightStep;
		center += leftRightStep;
	}
	if (currentKeyStates[SDL_SCANCODE_D]) {
		position -= leftRightStep;
		center -= leftRightStep;
	}
	if (currentKeyStates[SDL_SCANCODE_DOWN]) {
		position.y -= velocity;
		center.y -= velocity;
	}
	if (currentKeyStates[SDL_SCANCODE_UP]) {
		position.y += velocity;
		center.y += velocity;
	}

    //this->updateViewFrustum();
}

void Camera::lookAt(float dx, float dy) {
	float angleLR = -dx;
	float angleUD = dy;
	glm::vec3 normal = glm::cross(position - center, glm::vec3(0.0f, 1.0f, 0.0f));
	glm::mat4 transformation;
	transformation = glm::translate(transformation, position);
	transformation = glm::rotate(transformation, glm::radians(angleLR), glm::vec3(0.0f, 1.0f, 0.0f));
	if (normal != glm::vec3(0.0f, 0.0f, 0.0f)) {
		transformation = glm::rotate(transformation, glm::radians(angleUD), normal);
	}
	transformation = glm::translate(transformation, -position);
	glm::vec4 tempCenter = transformation * glm::vec4(center, 1.0f);
	center = glm::vec3(tempCenter.x, tempCenter.y, tempCenter.z);
}

glm::mat4 Camera::getViewMatrix() {
    return glm::lookAt(this->getPosition(), center, up);
}

glm::mat4 Camera::getProjectionMatrix() {
    return projectionMatrix;
}

glm::vec3 Camera::getViewDirection(){
	return glm::normalize(center - this->getPosition());
}

glm::vec3 Camera::getPosition() {
    return position;
}

void Camera::setProjectionMatrix(float angle, float ratio, float nearPlaneDist, float farPlaneDist) {
    this->frustumAngle = angle;
    this->ratio = ratio;
    this->nearPlaneDistance = nearPlaneDist;
    this->farPlaneDistance = farPlaneDist;

    float tang = glm::tan(M_PI / 180.0f * angle * 0.5);
    this->nearPlaneHeight = nearPlaneDist * tang;
    this->nearPlaneWidth = this->nearPlaneHeight * ratio;
    this->farPlaneHeight = farPlaneDist * tang;
    this->farPlaneWidth = this->farPlaneHeight * ratio;

    this->projectionMatrix = glm::perspective(glm::radians(angle), ratio, nearPlaneDist, farPlaneDist);
}

/*
void Camera::updateViewFrustum() {
    frustumPlanes.clear();

    glm::vec3 cameraZAxis = -this->getViewDirection();
    glm::vec3 cameraXAxis = glm::normalize(glm::cross(up, cameraZAxis));
    glm::vec3 cameraUpVector = glm::cross(cameraZAxis, cameraXAxis);

    glm::vec3 farPlaneCenter = position - cameraZAxis * farPlaneDistance;
    glm::vec3 nearPlaneCenter = position - cameraZAxis * nearPlaneDistance;

    PlaneInfo farPlane;
    farPlane.plane = cameraZAxis;
    farPlane.pointOnPlane = farPlaneCenter;
    PlaneInfo nearPlane;
    nearPlane.plane = -cameraZAxis;
    nearPlane.pointOnPlane = nearPlaneCenter;

    PlaneInfo topPlane;
    glm::vec3 directionHelper = glm::normalize((farPlaneCenter + cameraUpVector * farPlaneHeight) - this->position);
    topPlane.plane = glm::cross(directionHelper, cameraXAxis);
    topPlane.pointOnPlane = farPlaneCenter + cameraUpVector * farPlaneHeight;
    PlaneInfo bottomPlane;
    directionHelper = glm::normalize((farPlaneCenter - cameraUpVector * farPlaneHeight) - this->position);
    bottomPlane.plane = glm::cross(cameraXAxis, directionHelper);
    bottomPlane.pointOnPlane = farPlaneCenter - cameraUpVector * farPlaneHeight;

    PlaneInfo leftPlane;
    directionHelper = glm::normalize((farPlaneCenter - cameraXAxis * farPlaneWidth) - this->position);
    leftPlane.plane = glm::cross(directionHelper, cameraUpVector);
    leftPlane.pointOnPlane = farPlaneCenter - cameraXAxis * farPlaneWidth;
    PlaneInfo rightPlane;
    directionHelper = glm::normalize((farPlaneCenter + cameraXAxis * farPlaneWidth) - this->position);
    rightPlane.plane = glm::cross(cameraUpVector, directionHelper);
    rightPlane.pointOnPlane = farPlaneCenter + cameraXAxis * farPlaneWidth;

    frustumPlanes.push_back(farPlane);
    frustumPlanes.push_back(nearPlane);
    frustumPlanes.push_back(leftPlane);
    frustumPlanes.push_back(rightPlane);
    frustumPlanes.push_back(topPlane);
    frustumPlanes.push_back(bottomPlane);

}

float Camera::getPlanePointDistance(PlaneInfo planeInfo, glm::vec3 point) {
    return glm::dot(planeInfo.plane, (point - planeInfo.pointOnPlane));
}

bool Camera::isInViewFrustum(BoundingBox bb) {
    std::vector<glm::vec3> points;
    points.push_back(bb.min);
    points.push_back(bb.min + glm::vec3(bb.max.x - bb.min.x, 0, 0));
    points.push_back(bb.min + glm::vec3(0, 0, bb.max.z - bb.min.z));
    points.push_back(bb.min + glm::vec3(bb.max.x - bb.min.x, 0, bb.max.z - bb.min.z));
    points.push_back(bb.max);
    points.push_back(bb.min + glm::vec3(bb.max.x - bb.min.x, bb.max.y - bb.min.y, 0));
    points.push_back(bb.min + glm::vec3(0, bb.max.y - bb.min.y, bb.max.z - bb.min.z));
    points.push_back(bb.min + glm::vec3(0, bb.max.y - bb.min.y, 0));

    for (PlaneInfo p : frustumPlanes) {
        int pointsOnWrongSide = 0;
        for (glm::vec3 point : points) {
            if (getPlanePointDistance(p, point) < 0) {
                pointsOnWrongSide++;
            } else {
                break;
            }
        }
        if (pointsOnWrongSide >= points.size()) {
            return false;
        }
    }

    return true;
}*/